/****************************************************************************
**	File:	XNUT4.C
**
**	Desc:	Example of a NUT application.
**
**	Disclaimer:
**
**		Novell, Inc. makes no representations or warranties with respect to
**		any NetWare software, and specifically disclaims any express or
**		implied warranties of merchantability, title, or fitness for a
**		particular purpose.  
**
**		Distribution of any NetWare software is forbidden without the
**		express written consent of Novell, Inc.  Further, Novell reserves
**		the right to discontinue distribution of any NetWare software.
**
**		Novell is not responsible for lost profits or revenue, loss of use
**		of the software, loss of data, costs of re-creating lost data, the
**		cost of any substitute equipment or program, or claims by any party
**		other than you.  Novell strongly recommends a backup be made before
**		any software is installed.   Technical support for this software
**		may be provided at the discretion of Novell.
**
**
**	QMK386 options used:
**
**		/in - Include import statement for NWSNUT symbols.
**		/x  - No default screen for NLM.
**		/z  - Generate NLMLINKx version of def file.
**
**		NOTE:  /id (include NDS imports) is not needed with this NLM because
**		this NLM does not rely on the linker to import it's NDS symbols.
**		Rather, this NLM attempts to import NDS symbols at run-time using the
**		ImportSymbol() function.  Therefore, all NDS functions are called using
**		function pointers.
**
**	Programmers:
**
**		Ini	Who						Firm
**		-----------------------------------------------------------------------
**		ABJ	Adam B. Jerome			Novell Developer Support.
**
**	History:
**
**		When		Who	What
**		-----------------------------------------------------------------------
**		08-19-94	ABJ	First code.
**		11-08-95 ABJ	Updated code adding several features and enhancements.
**							(You can learn a lot in a year).
*/

/****************************************************************************
**	Compiler setup.
*/
	/*------------------------------------------------------------------------
	**	ANSI
	*/
	#include <stdlib.h>	/*	exit()	*/
	#include <signal.h>	/* SIGTERM, signal() */
	#include <stdio.h>	/* sprintf() */

	/*------------------------------------------------------------------------
	**	NetWare
	*/
	#define _SIZE_T_DEFINED 
   #include <nwdsapi.h>	/*	NWDSContextHandle(), NWCONN_HANDLE, MAX_DN_CHARS, NWDSCCODE, DCK_FLAGS, NWDSCreateContext(), NWDSGetContext() */
	#include <advanced.h>/* AllocatResourceTag() */
	#include <nwsnut.h>	/* NWSRestoreNut(), NWSPushList(), ...*/
   #include <nwconn.h> 	/*	GetMaximumNumberOfStations(), GetConnectionInformation()*/
   #include <nwcntask.h>/*	SetCurrentConnection(), GetCurrentConnection(), ReturnConnection() */
   #include <nwmisc.h> 	/* LongSwap() */
	#include <conio.h>  	/* CreateScreen(), DestroyScreen() */
	#include <process.h>	/* GetNLMHandle() */
	#include <nwenvrn.h>	/*	FILE_SERVER_INFO, GetServerInformation() */

	/*------------------------------------------------------------------------
	**	XNUT4
	*/
	#include "XNUT4.mlc"	/* File maintained by MSGLIB.EXE */
	#include "XNUT4.mlh"	/* File maintained by MSGLIB.EXE */

/****************************************************************************
**	Global storage
*/
	/*------------------------------------------------------------------------
	**	Variables
	*/
	int		NLM_exiting 	=	FALSE;
	int		NLM_threadCnt	=	0;
	int 		NLM_screenID	=	(int)NULL;
	NUTInfo *NLM_nutHandle 	=	NULL;
	int		NLM_ndsActive 	=	FALSE;

	int		NLM_mainThreadGroupID;

	/*------------------------------------------------------------------------
	**	NDS Function pointers
	*/
	NWDSContextHandle NWFAR NWPASCAL (*NLM_NWDSCreateContext)
		(void)	=	NULL;

	NWDSCCODE NWFAR NWPASCAL (*NLM_NWDSGetContext)
		(
		NWDSContextHandle	context,
		int	key,
		void 	NWFAR *value
		) = NULL;

	NWDSCCODE NWFAR NWPASCAL (*NLM_NWDSSetContext)
		(
		NWDSContextHandle	context,
		int	key,
		void 	NWFAR *value
		) = NULL;

	NWDSCCODE NWFAR NWPASCAL (*NLM_NWDSMapIDToName)
		(
		NWDSContextHandle		context,
		int		connID,
		uint32	objectID,
		char		NWFAR *object
		) = NULL;

	NWDSCCODE NWFAR NWPASCAL (*NLM_NWDSFreeContext)
		(
		NWDSContextHandle	context
		) = NULL;

	
/****************************************************************************
** Main menu action procedure.
*/
int NLM_VerifyProgramExit(void)
	{
	int cCode;
	
	cCode=NWSConfirm(
		/*	I-	header				*/	PROGRAM_EXIT,
		/*	I-	centerLine			*/	0,
		/*	I-	centerColumn		*/	0,
		/*	I-	defaultChoice		*/	1,
		/*	I-	actionProcedure	*/	NULL,
		/*	I-	handle				*/	NLM_nutHandle,
		/*	I-	actionParameter	*/	NULL
		);
	
	/*------------------------------------------------------------------------
	**	Escape(-1) means No(0).
	*/
	if(cCode == (-1))
		cCode = 0;

	return(cCode);
	}

/****************************************************************************
**	Display information about a specific connection.
*/
void NLM_DisplayConnectionInformation(char *text, WORD conn, int mode)
	{	
   char  objectName[47+1];
   WORD  objectType;
   long  objectID;
   BYTE  loginTime[7];
	int	cCode;

	LONG  portal;
	PCB  *portalPCB;

	char 	szTemp[80+1];

	/*------------------------------------------------------------------------
	**	Get info about connection.
	*/
	cCode=GetConnectionInformation(
		/* I- connectionNumber 	*/ conn,
		/*	I-	objectName			*/	objectName,
		/*	-O	objectType			*/	&objectType,
		/*	-O	objectID				*/	&objectID,
		/*	-O	loginTime			*/	loginTime
		);
	if(cCode != 0)
		{
		NWSTrace(NLM_nutHandle, "GetConnectionInformation(%d) reports: %d",
			conn,
			cCode
			);
		return;
		}

	/*------------------------------------------------------------------------
	** Dim the current portal	
	*/
	NWSDeselectPortal(NLM_nutHandle);

	/*------------------------------------------------------------------------
	**	Create a portal in which we will display the connection information.
	**	(A portal is a window).
	*/
	portal=NWSCreatePortal(
		/*	I-	line		  			*/ 3,
		/*	I-	column				*/	2,
		/*	I-	frameHeight			*/	10,
		/*	I-	frameWidth			*/	76,
		/*	I-	virtualHeight		*/	10,
		/*	I-	virtualWidth		*/	76,
		/*	I-	saveFlag				*/	SAVE,
		/*	I-	headerText			*/	text,
		/*	I-	headerAttribute	*/	VNORMAL,
		/*	I-	borderType			*/	SINGLE,
		/*	I-	borderAttribute	*/	VINTENSE,
		/*	I-	cursorFlag			*/	CURSOR_OFF,
		/*	I-	directflag			*/	VIRTUAL,
		/*	I-	handle				*/	NLM_nutHandle
		);
	switch(portal)
		{
		case 0xFFFFFFFE:
			NWSTrace(NLM_nutHandle, "NWSCreatePortal reports: Unable to "
				"allocate memory for PCB, virtual screen, or save area.");
			return;

		case 0xFFFFFFFF:
			NWSTrace(NLM_nutHandle, "NWSCreatePortal reports: Maximum "
				"number of portals already defined.");
			return;
		}

	/*------------------------------------------------------------------------
	**	Get portal's PCB.
	*/
	NWSGetPCB(&portalPCB, portal, NLM_nutHandle);

	/*------------------------------------------------------------------------
	**	Make our portal current and clear it.
	*/
	NWSSelectPortal(portal, NLM_nutHandle);
	NWSClearPortal(portalPCB);

	/*------------------------------------------------------------------------
	**	Place information on portal.
	*/
	NWSDisplayTextInPortal(0,0,"<<Press ESCAPE to exit>>",VINTENSE,portalPCB);
	
	sprintf(szTemp, "List mode: %s", (mode == MENU_MAIN_CON_BIND) ? "BINDERY" : "NDS"); 
	NWSDisplayTextInPortal(2,0,szTemp,VINTENSE,portalPCB);

	sprintf(szTemp, "Object Type: 0x%04X", objectType); 
	NWSDisplayTextInPortal(4,0,szTemp,VNORMAL,portalPCB);

	sprintf(szTemp, "Object ID..: 0x%08X", objectID); 
	NWSDisplayTextInPortal(5,0,szTemp,VNORMAL,portalPCB);

	/*------------------------------------------------------------------------
	**	Update portal content to user screen.
	*/
	NWSUpdatePortal(portalPCB);

	/*------------------------------------------------------------------------
	**	Wait for user to press ESCAPE.
	*/
	NWSWaitForEscape(NLM_nutHandle);

	/*------------------------------------------------------------------------
	**	Trash portal.
	*/
	NWSDestroyPortal(
		/* I-	portalNumber	*/	portal,
		/*	I-	handle			*/ NLM_nutHandle
		);

	return;
	}

/****************************************************************************
** Action procedure for both NDS and BINDERY connection lists.
*/
int NLM_ListSubAct(LONG keyPressed, LIST **elementSelected,
		LONG *itemLineNumber, void *actionParameter)
	{

	actionParameter=actionParameter;		/* Rid compiler warning */
	itemLineNumber=itemLineNumber;		/* Rid compiler warning */

	switch(keyPressed)
		{
		case M_ESCAPE:
			return(0);

		case M_SELECT:
			NLM_DisplayConnectionInformation(
				/*	I-	ListText					*/	(*elementSelected)->text,
				/* I-	Connection Number 	*/ *(WORD *)((*elementSelected)->otherInfo),
				/*	I- Mode (DS or Bindery 	*/	(int)actionParameter
				);
			break;
		}

	return(-1);
	}

/****************************************************************************
**	Sub-list build procedure - Bindery version.
**
**	The only function in this routine which accesses the NWSNUT library is the
**	"NWSAppendToList()" function.  Each time NWSAppendToList() is called, it
** will add a new (linked-list) node to the current (active) list.
*/
void NLM_ListSubBuildBindery(void)
	{
   WORD  connectionNumber;
   char  objectName[47+1];
   WORD  objectType;
   long  objectID;
   BYTE  loginTime[7];
   int 	cCode;
	int	connMax;
	char	szTemp[80+1];
	char  network[4];
	char  node[6];
	WORD	*wp;

   /*------------------------------------------------------------------------
   ** Examine each connection
   */
	connMax=GetMaximumNumberOfStations();
   for(connectionNumber=0; connectionNumber < connMax; connectionNumber++)
   	{
		if(NLM_exiting)
			break;

		/*---------------------------------------------------------------------
		**	Get info about connection and verify that it is in use.
		*/
		cCode=GetConnectionInformation(
			/* I- connectionNumber 	*/ connectionNumber,
			/*	I-	objectName			*/	objectName,
			/*	-O	objectType			*/	&objectType,
			/*	-O	objectID				*/	&objectID,
			/*	-O	loginTime			*/	loginTime
			);

		/*---------------------------------------------------------------------
		**	Build a displayable string for the list, and append it to the list.
		*/
		if(cCode != 0)
			sprintf(szTemp, "%4d GetConnectionInformation() = ERROR:%d", connectionNumber, cCode);
		else if(objectID == 0) 	
			{
			cCode=GetInternetAddress(
				/* I-	connectionNumber	*/	connectionNumber,
				/*	-O networkNumber		*/ network,
				/*	-O	nodeAddress			*/	node
				);
 			if(cCode != 0)
				sprintf(szTemp, "%4d Not-In-Use", connectionNumber);
			else
				sprintf(szTemp, "%4d Not-Logged-In %02X%02X%02X%02X:%02X%02X%02X%02X%02X%02X", 
				connectionNumber,
				network[0],
				network[1],
				network[2],
				network[3],
				node[0],
				node[1],
				node[2],
				node[3],
				node[4],
				node[5]
				);
			}
		else
			sprintf(szTemp, "%4d %.47s", connectionNumber, objectName);

		/*---------------------------------------------------------------------
		**	Allocate storage in linked-list node for the node's connection number.
		*/
		wp=(WORD *)malloc(sizeof(WORD));
		if(wp == NULL)
			{
			NWSTrace(NLM_nutHandle, "WARNING: Out of memory.  List is not complete.");
			break;
			}
		*wp=connectionNumber;

		NWSAppendToList(szTemp, wp, NLM_nutHandle);
   	}

	return;
	}
										  
/****************************************************************************
**	Sub-list build procedure - NDS version.
**
** Desc: From XDSA4.EXE.  Lists users on a server and displays their full
**			NDS name.
**
** Programmers:
**
**    Ini   Who                  Firm
**    ---------------------------------------------------------------------
**    DWH   Dirk W. Howard       Novell Developer Support
**		ABJ	Adam B. Jerome			Novell Developer Support
**
**
** History:
**
**    When    	Who   What
**    ---------------------------------------------------------------------
**    11-03-95 DWH	First code.
**		11-08-95	ABJ	Snatched and modified code from XDSA4.EXE.
*/
void NLM_ListSubBuildNDS(void)
	{
   WORD                 connectionNumber;
   char                 objectName[47+1];
   WORD                 objectType;
   long                 objectID;
   BYTE                 loginTime[7];
   NWDSContextHandle    cx;
   NWCONN_HANDLE        connID;
   char                 name[MAX_DN_CHARS+1];
   NWDSCCODE            cCode;
   LONG	               flags;
	int						connMax;
	char						szTemp[80+1];
	char                 network[4];
	char                 node[6];
	WORD					  *wp;

   /*------------------------------------------------------------------------
   ** Create a DS context to work with	and get and set the context flags so
	**	we can display the names in a typeless naming convention and the full
	**	name is used.
   */
   cx = (*NLM_NWDSCreateContext)();
   cCode = (*NLM_NWDSGetContext)(cx, DCK_FLAGS, &flags);
   if(cCode)
      goto END_ERR;

   flags |= DCV_TYPELESS_NAMES;        /* Set the Typeless names flag */
   flags &= ~DCV_CANONICALIZE_NAMES;   /* Clear the Canonicalize names flag */

   cCode = (*NLM_NWDSSetContext)(cx, DCK_FLAGS, &flags);
   if(cCode)
   	goto END_ERR;

   /*------------------------------------------------------------------------
   ** Get a connection other than connection 0 to work on.
	**	(Because NWDS functions require something other than conn zero).
   */
   SetCurrentConnection(-1);
   connID = GetCurrentConnection();

   /*------------------------------------------------------------------------
   ** Examine each connection
   */
	connMax=GetMaximumNumberOfStations();
   for(connectionNumber=0; connectionNumber < connMax; connectionNumber++)
   	{
		if(NLM_exiting)
			break;

		/*---------------------------------------------------------------------
		**	Get info about connection and verify that it is in use.
		*/
		cCode=GetConnectionInformation(
			/* I-	connectionNumber 	*/ connectionNumber,
			/*	-O	objectName			*/	objectName,
			/*	-O	objectType			*/	&objectType,
			/*	-O	objectID				*/	&objectID,
			/*	-O	loginTime			*/	loginTime
			);

		/*---------------------------------------------------------------------
		**	Build a displayable string for the list, and append it to the list.
		*/
		if(cCode != 0)
			sprintf(szTemp, "%4d GetConnectionInformation() = ERROR:%d", connectionNumber, cCode);
		else if(objectID == 0) 	
			{
			cCode=GetInternetAddress(
				/* I-	connectionNumber	*/	connectionNumber,
				/*	-O networkNumber		*/ network,
				/*	-O	nodeAddress			*/	node
				);
 			if(cCode != 0)
				sprintf(szTemp, "%4d Not-In-Use", connectionNumber);
			else
				sprintf(szTemp, "%4d Not-Logged-In %02X%02X%02X%02X:%02X%02X%02X%02X%02X%02X %s", 
				connectionNumber,
				network[0],
				network[1],
				network[2],
				network[3],
				node[0],
				node[1],
				node[2],
				node[3],
				node[4],
				node[5],
				(connID == connectionNumber) ? "[This NLM's connection number]" : " "
				);
			}
		else
			{
      	(*NLM_NWDSMapIDToName)(cx, connID, LongSwap(objectID), name);
			sprintf(szTemp, "%4d %.60s", connectionNumber, name);
			}

		/*---------------------------------------------------------------------
		**	Allocate storage in linked-list node for the node's connection number.
		*/
		wp=(WORD *)malloc(sizeof(WORD));
		if(wp == NULL)
			{
			NWSTrace(NLM_nutHandle, "WARNING: Out of memory.  List is not complete.");
			break;
			}
		*wp=connectionNumber;

		NWSAppendToList(szTemp, wp, NLM_nutHandle);
   	}

   /*------------------------------------------------------------------------
   ** Return connection
   */
   ReturnConnection(SetCurrentConnection(0));

END_ERR:

   /*------------------------------------------------------------------------
   ** Return the context
   */
   (*NLM_NWDSFreeContext)(cx);

	return;
	}
										  
/****************************************************************************
** Sub-list
*/
void NLM_ListSub(int mode)
	{
	/*------------------------------------------------------------------------
	**	Don't do this list if we should be exiting.
	*/
	if(NLM_exiting)
		return;

	/*------------------------------------------------------------------------
	**	At this point, the current list is the Main Menu.  If we begin adding
	**	new items to the current list, it would mess up the Main menu (to say
	**	the least).  So, we will save the Main Menu List on the List stack
	** (PushList) and then initialize a new list (set head and tail to NULL)
	**	by calling InitList().  Note that Lists use NWInitList() and Menus use
	**	NWInitMenu().
	*/
	NWSPushList(NLM_nutHandle);
	NWSInitList(NLM_nutHandle, free);

	/*------------------------------------------------------------------------
	**	This one function (NLM_ListSub) is called for both an NDS and Bindery
	**	style list.  The list that will actually be built depends on the value
	** of mode.
	*/
	if(mode == MENU_MAIN_CON_NDS)
		NLM_ListSubBuildNDS();
	else
		NLM_ListSubBuildBindery();

	/*------------------------------------------------------------------------
	**	Display the list and allow user interaction.
	*/
	NWSList(
		/* I-	header				*/	LIST_SUBLIST__HDR,
		/*	I-	centerLine			*/	0,
		/*	I-	centerColumn		*/	0,
		/*	I-	height				*/	10,
		/* I-	width					*/ 72,
		/* I-	validKeyFlags		*/	M_ESCAPE|M_SELECT,
		/*	IO	element				*/	NULL,
		/*	I-	handle				*/ NLM_nutHandle,
		/*	I-	formatProcedure	*/	NULL,
		/* I-	actionProcedure	*/	NLM_ListSubAct,
		/* I-	actionParameter	*/	(void *)mode
		);
	
	/*------------------------------------------------------------------------
	**	Before returning, we must free the list items allocated by
	**	NLM_ListSubBuild...().  Then the Main Menu list context
	**	must be restored.  Note that Lists use NWDestroyList() and
	** Menus use NWDestroyMenu().
	*/
	NWSDestroyList(NLM_nutHandle);
	NWSPopList(NLM_nutHandle);

	return;
	}

/****************************************************************************
** Unsorted sub-menu (NWSLIST) action procedure.  Note that the parameters
**	for an NWSList() action procedure are very different from the parameters
**	passed to an NWSMenu() action procedure.
*/
int NLM_MenuNoSortAct(LONG keyPressed, LIST **elementSelected,
		LONG *itemLineNumber, void *actionParameter)
	{
	int index;

	actionParameter=actionParameter;		/* Rid compiler warning */
	itemLineNumber=itemLineNumber;		/* Rid compiler warning */

	/*------------------------------------------------------------------------
	**	Setup index variable to be the same as it would be in a NWSMenu()
	**	action procedure.
	*/
	if(keyPressed == M_ESCAPE)
		index = (-1);
	else
		index=*(int *)((*elementSelected)->otherInfo);

	/*------------------------------------------------------------------------
	**	Perform the user-selected action.
	**	(Just like a normal NWSMenu() action procedure...)
	*/
	switch(index)
		{
		case (-1): 		/* ESCAPE KEY PRESSED */
			return(0);

		case MENU_NOSORT_OPTION_A:
			NWSTrace(NLM_nutHandle, "Insert no-sort sub-menu option #A here.");
			break;

		case MENU_NOSORT_OPTION_B:
			NWSTrace(NLM_nutHandle, "Insert no-sort sub-menu option #B here.");
			break;

		case MENU_NOSORT_OPTION_C:
			NWSTrace(NLM_nutHandle, "Insert no-sort sub-menu option #C here.");
			break;

		case MENU_NOSORT_OPTION_D:
			NWSTrace(NLM_nutHandle, "Insert no-sort sub-menu option #D here.");
			break;

		case MENU_NOSORT_OPTION_E:
			NWSTrace(NLM_nutHandle, "Insert no-sort sub-menu option #E here.");
			break;

		default:
			NWSTrace(NLM_nutHandle, "Option not implemented.");
			break;
		}

	/*------------------------------------------------------------------------
	**	If we should be exiting, pretend that ESCAPE was pressed.
	*/
	if(NLM_exiting)
		return(0);

	return(-1);
	}

/****************************************************************************
** Unsorted sub-menu.
**
** There are times when you would like to display a menu, but you don't want
**	the elements to be sorted.  NWSMenu() automatically sorts the list of menu
**	items and there is no way to disable this feature.
**
** The NWSList() function has an M_NO_SORT flag that is not available to the
**	NWSMenu() function; however, using NWSList to display a menu can be scary
**	if you don't know how.
**
**	The following code demonstrates how to build a menu and then display it as
** a list.  The action procedure (above) is specific to NWSList() and is not
**	a suitable action procedure for NWSMenu().
*/
void NLM_MenuNoSort(void)
	{
	LIST *d;		/* Used to specify the "default" menu item. */

	/*------------------------------------------------------------------------
	**	Don't do this menu if we should be exiting.
	*/
	if(NLM_exiting)
		return;

	/*------------------------------------------------------------------------
	**	At this point, the current list is the Main Menu.  If we begin adding
	**	new items to the current list, it would mess up the Main menu (to say
	**	the least).  So, we will save the Main Menu List on the List stack
	** (PushList) and then initialize a new list (set head and tail to NULL)
	**	by calling InitMenu().  Note that Lists use NWInitList() and Menus use
	**	NWInitMenu().
	*/
	NWSPushList(NLM_nutHandle);
	NWSInitMenu(NLM_nutHandle);

	/*------------------------------------------------------------------------
	**	Insert menu items in the order they will be displayed.
	*/
	NWSAppendToMenu(MENU_NOSORT_OPTION_B, MENU_NOSORT_OPTION_B, NLM_nutHandle);
	NWSAppendToMenu(MENU_NOSORT_OPTION_A, MENU_NOSORT_OPTION_A, NLM_nutHandle);
	d=NWSAppendToMenu(MENU_NOSORT_OPTION_C, MENU_NOSORT_OPTION_C, NLM_nutHandle);
	NWSAppendToMenu(MENU_NOSORT_OPTION_E, MENU_NOSORT_OPTION_E, NLM_nutHandle);
	NWSAppendToMenu(MENU_NOSORT_OPTION_D, MENU_NOSORT_OPTION_D, NLM_nutHandle);

	/*------------------------------------------------------------------------
	**	Display the menu (as though it were a list) and allow user interaction.
	*/
	NWSList(
		/* header				*/	MENU_NOSORT__HDR,
		/*	centerLine			*/	0,
		/*	centerColumn		*/	65,
		/*	height				*/	5,
		/* width					*/ 20,
		/* validKeyFlags		*/	M_ESCAPE|M_SELECT|M_NO_SORT,
		/*	element				*/	&d,
		/*	handle				*/ NLM_nutHandle,
		/*	formatProcedure	*/	NULL,
		/* actionProcedure	*/	NLM_MenuNoSortAct,
		/* actionParameter	*/	NULL
		);
	
	/*------------------------------------------------------------------------
	**	Before returning, we must free the list items allocated by
	**	NWSAppendToMenu().  Then the Main Menu list context must be restored.
	**	Note that Lists use NWDestroyList() and Menus use NWDestroyMenu().
	*/
	NWSDestroyMenu(NLM_nutHandle);
	NWSPopList(NLM_nutHandle);

	return;
	}

/****************************************************************************
** Sub menu action procedure.
*/
int NLM_MenuSubAct(int index, void *parm)
	{
	parm=parm;	 /* Rid compiler warning. */

	/*------------------------------------------------------------------------
	**	Perform the user-selected action.
	*/
	switch(index)
		{
		case (-1): 		/* ESCAPE KEY PRESSED */
			return(0);

		case MENU_SUB_OPTION1:
			NWSTrace(NLM_nutHandle, "Insert sub-menu option #1 here.");
			break;

		case MENU_SUB_OPTION2:
			NWSTrace(NLM_nutHandle, "Insert sub-menu option #2 here.");
			break;

		default:
			NWSTrace(NLM_nutHandle, "Option not implemented.");
			break;
		}

	/*------------------------------------------------------------------------
	**	If we should be exiting, pretend that ESCAPE was pressed.
	*/
	if(NLM_exiting)
		return(0);

	return(-1);
	}

/****************************************************************************
** Sub menu.
*/
void NLM_MenuSub(void)
	{
	/*------------------------------------------------------------------------
	**	Don't do this menu if we should be exiting.
	*/
	if(NLM_exiting)
		return;

	/*------------------------------------------------------------------------
	**	At this point, the current list is the Main Menu.  If we begin adding
	**	new items to the current list, it would mess up the Main menu (to say
	**	the least).  So, we will save the Main Menu List on the List stack
	** (PushList) and then initialize a new list (set head and tail to NULL)
	**	by calling InitMenu().  Note that Lists use NWInitList() and Menus use
	**	NWInitMenu().
	*/
	NWSPushList(NLM_nutHandle);
	NWSInitMenu(NLM_nutHandle);

	/*------------------------------------------------------------------------
	**	Insert menu items.  Note that the insertion order does not matter being
	**	that NWSMenu() will always sort the Menu selections automatically.
	*/
	NWSAppendToMenu(MENU_SUB_OPTION1, MENU_SUB_OPTION1, NLM_nutHandle);
	NWSAppendToMenu(MENU_SUB_OPTION2, MENU_SUB_OPTION2, NLM_nutHandle);

	/*------------------------------------------------------------------------
	**	Display the menu and allow user interaction.
	*/
	NWSMenu(
		/*	header				*/	MENU_SUB__HDR,
		/*	centerLine			*/	0,					
		/*	centerColumn		*/	15,
		/*	defaultElement		*/	NULL,
		/*	actionProcedure	*/	NLM_MenuSubAct,
		/*	handle				*/	NLM_nutHandle,
		/*	actionParameter	*/	NULL
		);
	
	/*------------------------------------------------------------------------
	**	Before returning, we must free the list items allocated by
	**	NWSAppendToMenu().  Then the Main Menu list context must be restored.
	**	Note that Lists use NWDestroyList() and Menus use NWDestroyMenu().
	*/
	NWSDestroyMenu(NLM_nutHandle);
	NWSPopList(NLM_nutHandle);

	return;
	}

/****************************************************************************
** Main menu action procedure.
*/
int NLM_MenuMainAct(int index, void *parm)
	{
	parm=parm;	 /* Rid compiler warning. */

	/*------------------------------------------------------------------------
	**	Perform the user-selected action.
	*/
	switch(index)
		{
		case (-1): 		/* ESCAPE KEY PRESSED */
			if(NLM_VerifyProgramExit())
				return(0);

			break;

		case MENU_MAIN_SUBMENU:
			NLM_MenuSub();
			break;

		case MENU_MAIN_NOSORT:
			NLM_MenuNoSort();
			break;

		case MENU_MAIN_CON_BIND:
		case MENU_MAIN_CON_NDS:
			NLM_ListSub(index);
			break;

		default:
			NWSTrace(NLM_nutHandle, "Option not implemented.");
			break;
		}

	/*------------------------------------------------------------------------
	**	If we should be exiting, pretend that ESCAPE was pressed.
	*/
	if(NLM_exiting)
		return(0);

	return(-1);
	}

/****************************************************************************
** Main menu.
*/
void NLM_MenuMain(void)
	{
	LIST *defaultOption;

	/*------------------------------------------------------------------------
	**	Don't do this menu if we should be exiting.
	*/
	if(NLM_exiting)
		return;

	/*------------------------------------------------------------------------
	**	At this point, the current list is uninitialized (being that it is the
	** first list of the program.)  Before using the current list it must be
	**	initialized (set head and tail to NULL) by calling InitMenu().
	**	Note that Lists use NWInitList() and Menus use NWInitMenu().
	*/
	NWSInitMenu(NLM_nutHandle);

	/*------------------------------------------------------------------------
	**	Insert menu items.  Note that the insertion order does not matter being
	**	that NWSMenu() will always sort the Menu selections automatically.
	** The defaultOption stores a pointer to the menu item which we wish to be
	** highlighed by default.
	*/
	defaultOption =
		NWSAppendToMenu(MENU_MAIN_SUBMENU, MENU_MAIN_SUBMENU, NLM_nutHandle);

	NWSAppendToMenu(MENU_MAIN_NOSORT, MENU_MAIN_NOSORT, NLM_nutHandle);
	NWSAppendToMenu(MENU_MAIN_CON_BIND, MENU_MAIN_CON_BIND, NLM_nutHandle);
	
	/* This menu item only appears if DSAPI.NLM has been loaded. */
	if(NLM_ndsActive)
		NWSAppendToMenu(MENU_MAIN_CON_NDS, MENU_MAIN_CON_NDS, NLM_nutHandle);

	/*------------------------------------------------------------------------
	**	Display the menu and allow user interaction.
	*/
	NWSMenu(
		/*	header				*/	MENU_MAIN__HDR,
		/*	centerLine			*/	0,					/* [0,0] means Center of screen*/
		/*	centerColumn		*/	0,
		/*	defaultElement		*/	defaultOption,	/* Could use NULL here desired */			
		/*	actionProcedure	*/	NLM_MenuMainAct,
		/*	handle				*/	NLM_nutHandle,
		/*	actionParameter	*/	NULL
		);
	
	/*------------------------------------------------------------------------
	**	Before returning, we must free the list items allocated by
	**	NWSAppendToMenu(). Note that Lists use NWDestroyList() and Menus use
	**	NWDestroyMenu().
	*/
	NWSDestroyMenu(NLM_nutHandle);

	return;
	}

/****************************************************************************
** Program shut-down.
*/
void NLM_ShutDown(void)
	{

	/*------------------------------------------------------------------------
	**	Shut down NWSNUT.
	*/
	if(NLM_nutHandle != NULL)
		NWSRestoreNut(NLM_nutHandle);

	/*------------------------------------------------------------------------
	**	We won't need this screen any more.
	*/
	if(NLM_screenID != NULL)
		DestroyScreen(NLM_screenID);
	
	/*------------------------------------------------------------------------
	**	Inform DSAPI.NLM that we no longer require his services
	**	(as appropriate).
	*/
	if(NLM_NWDSCreateContext != NULL)
		UnimportSymbol(GetNLMHandle(), "NWDSCreateContext");

	if(NLM_NWDSGetContext != NULL)
		UnimportSymbol(GetNLMHandle(), "NWDSGetContext");

	if(NLM_NWDSSetContext != NULL)
		UnimportSymbol(GetNLMHandle(), "NWDSSetContext");
	
	if(NLM_NWDSMapIDToName != NULL)
		UnimportSymbol(GetNLMHandle(), "NWDSMapIDToName");

	if(NLM_NWDSFreeContext != NULL)
		UnimportSymbol(GetNLMHandle(), "FreeContext");

	return;
	}

/****************************************************************************
** Program signal handler.
*/
void NLM_SignalHandler(int sig)
	{
	int handlerThreadGroupID;
	time_t t;

	switch(sig)
		{
  		/*---------------------------------------------------------------------
		**	SIGTERM is called when this NLM is unloaded via console command.
		*/
		case SIGTERM:

			/*------------------------------------------------------------------
			**	Tell all of our threads to "get out of the pool."
			*/
			NLM_exiting=TRUE;

			/*------------------------------------------------------------------
			** Wait for main() to terminate.
			**
			** If main() has not terminateded within a 1/2 second, ungetch an
			**	escape key.  This will "trick" a blocking NWSList() or NWSMenu()
			**	function and wake it up. 
			*/
			handlerThreadGroupID = SetThreadGroupID(NLM_mainThreadGroupID);

			t=time(NULL);
			while(NLM_threadCnt != 0)
				{

				if(difftime(time(NULL), t) > 0.5)
					{
					ungetch(27);
					t=time(NULL);
					}

				ThreadSwitchWithDelay();
				}

			SetThreadGroupID(handlerThreadGroupID);

			break;

		}

	return;
	}

/****************************************************************************
** Import all DSAPI symbols our self so that we remain compatable with
**	NetWare 3.x.
*/
void NLM_ImportDSSymbols(void)
	{
	int cCode;
	FILE_SERV_INFO si;

	/*------------------------------------------------------------------------
	**	Bindery or NDS?
	*/
	cCode=GetServerInformation(
		/*	structSize	*/	sizeof(si),
		/*	serverInfo	*/	&si
		);
	if(cCode != 0)
		{
		NLM_exiting = TRUE;
		goto BINDERY;
		}
	if(si.netwareVersion < 4)
		goto BINDERY;

	/*------------------------------------------------------------------------
	**	Import DSAPI symbols.
	*/
	NLM_NWDSCreateContext = ImportSymbol(GetNLMHandle(), "NWDSCreateContext");
	if(NLM_NWDSCreateContext == NULL)
		goto BINDERY_OOPS;
		
	NLM_NWDSGetContext = ImportSymbol(GetNLMHandle(), "NWDSGetContext");
	if(NLM_NWDSGetContext == NULL)
		goto BINDERY_OOPS;
		
	NLM_NWDSSetContext = ImportSymbol(GetNLMHandle(), "NWDSSetContext"); 
	if(NLM_NWDSSetContext == NULL)
		goto BINDERY_OOPS;
		
	NLM_NWDSMapIDToName = ImportSymbol(GetNLMHandle(), "NWDSMapIDToName");
	if(NLM_NWDSMapIDToName == NULL)
		goto BINDERY_OOPS;

	NLM_NWDSFreeContext = ImportSymbol(GetNLMHandle(), "NWDSFreeContext");
	if(NLM_NWDSFreeContext == NULL)
		goto BINDERY_OOPS;

	NLM_ndsActive = TRUE;
	return;

BINDERY_OOPS:
	NWSTrace(NLM_nutHandle, "XNUT4 can use Directory Services calls if DSAPI.NLM is loaded previously.");

BINDERY:
	NLM_ndsActive = FALSE;

	return;
	}

/****************************************************************************
** Program initialization.
*/
void NLM_Initialize(void)
	{
	LONG	cCode;
	int	NLMHandle;
	LONG	tagID;

	/*------------------------------------------------------------------------
	**	Register SIGTERM hanler to facilitate a console UNLOAD command.
	*/
	signal(SIGTERM, NLM_SignalHandler);

	/*------------------------------------------------------------------------
	**	Get misc. values.
	*/
	NLMHandle=GetNLMHandle();							/* Needed by AllocateResourceTag() */
	NLM_mainThreadGroupID = GetThreadGroupID();	/* Needed by our SIGTERM handler   */

	/*------------------------------------------------------------------------
	**	Build our own screen.  This is not required.  I could have used just
	**	had the NLM use an automatically created screen, but I set the linker
	** option to screen="NONE" (QMK386 /x option) and included this code so
	**	that you could see how to create your own screen.
	*/
	NLM_screenID=CreateScreen(
		/*	screenName			*/	programMesgTable[SCREEN_NAME],
		/*	attributes			*/	AUTO_DESTROY_SCREEN
		);
	if(NLM_screenID == NULL)
		{
		NLM_exiting = TRUE;
		goto END;
		}

	/*------------------------------------------------------------------------
	**	Fire up NWSNUT on our screen.
	*/
	tagID=AllocateResourceTag(
		/*	NLMHandle			*/	NLMHandle,
		/*	descriptionString	*/	programMesgTable[RS_TAG_NAME],
		/*	resourceType		*/	AllocSignature
		);
	if(tagID == NULL)
		{
		NLM_exiting = TRUE;
		goto END;
		}

	cCode=NWSInitializeNut(
		/*	utility				*/	PROGRAM_NAME,
		/*	version				*/	PROGRAM_VERSION,
		/*	headerType			*/	NORMAL_HEADER,
		/*	compatibilityType	*/	NUT_REVISION_LEVEL,
		/*	messageTable		*/	programMesgTable,
		/*	helpScreens			*/	NULL,
		/*	screenID				*/	NLM_screenID,
		/*	resourceTag			*/	tagID,
		/*	handle				*/	&NLM_nutHandle
		);
	if(cCode != NULL)
		{
		NLM_exiting = TRUE;
		goto END;
		}

	/*------------------------------------------------------------------------
	**	Set our screen as current & active.
	*/
	DisplayScreen(NLM_screenID);

	/*------------------------------------------------------------------------
	**	Initialize our DS function pointers (if DSAPI.NLM is loaded).
	*/
	NLM_ImportDSSymbols();

END:
	
	return;
	}

/****************************************************************************
** Program start.
*/
void main(void)
	{
	++NLM_threadCnt;

	NLM_Initialize();
	NLM_MenuMain();
	NLM_ShutDown();
	
	--NLM_threadCnt;
	return;
	}



